@@ -1,32 +1,5 @@
-///////////////////////////////////////////////////////////////////////////
-//
-// Copyright (c) 2012-2017 DreamWorks Animation LLC
-//
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
-//
-// Redistributions of source code must retain the above copyright
-// and license notice and the following restrictions and disclaimer.
-//
-// *     Neither the name of DreamWorks Animation nor the names of
-// its contributors may be used to endorse or promote products derived
-// from this software without specific prior written permission.
-//
-// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY INDIRECT, INCIDENTAL,
-// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-// IN NO EVENT SHALL THE COPYRIGHT HOLDERS' AND CONTRIBUTORS' AGGREGATE
-// LIABILITY FOR ALL CLAIMS REGARDLESS OF THEIR BASIS EXCEED US$250.00.
-//
-///////////////////////////////////////////////////////////////////////////
+// Copyright Contributors to the OpenVDB Project
+// SPDX-License-Identifier: MPL-2.0
 //
 /// @file Math.h
 /// @brief General-purpose arithmetic and comparison routines, most of which
@@ -75,6 +48,16 @@
     #define OPENVDB_NO_FP_EQUALITY_WARNING_END
 #endif
 
+
+#ifdef OPENVDB_IS_POD
+#undef OPENVDB_IS_POD
+#endif
+#define OPENVDB_IS_POD(Type) \
+static_assert(std::is_standard_layout<Type>::value, \
+    #Type" must be a POD type (satisfy StandardLayoutType.)"); \
+static_assert(std::is_trivial<Type>::value, \
+    #Type" must be a POD type (satisfy TrivialType.)");
+
 namespace openvdb {
 OPENVDB_USE_VERSION_NAMESPACE
 namespace OPENVDB_VERSION_NAME {
@@ -89,6 +72,8 @@ template<> inline std::string zeroVal<std::string>() { return ""; }
 /// Return the @c bool value that corresponds to zero.
 template<> inline bool zeroVal<bool>() { return false; }
 
+namespace math {
+
 /// @todo These won't be needed if we eliminate StringGrids.
 //@{
 /// @brief Needed to support the <tt>(zeroVal<ValueType>() + val)</tt> idiom
@@ -99,13 +84,58 @@ inline std::string operator+(const std::string& s, float) { return s; }
 inline std::string operator+(const std::string& s, double) { return s; }
 //@}
 
+/// @brief  Componentwise adder for POD types.
+template<typename Type1, typename Type2>
+inline auto cwiseAdd(const Type1& v, const Type2 s)
+{
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_BEGIN
+    return v + s;
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_END
+}
+
+/// @brief  Componentwise less than for POD types.
+template<typename Type1, typename Type2>
+inline bool cwiseLessThan(const Type1& a, const Type2& b)
+{
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_BEGIN
+    return a < b;
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_END
+}
+
+/// @brief  Componentwise greater than for POD types.
+template<typename Type1, typename Type2>
+inline bool cwiseGreaterThan(const Type1& a, const Type2& b)
+{
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_BEGIN
+    return a > b;
+    OPENVDB_NO_TYPE_CONVERSION_WARNING_END
+}
+
+
+
+/// @brief  Pi constant taken from Boost to match old behaviour
+/// @note   Available in C++20
+template <typename T> inline constexpr T pi() { return 3.141592653589793238462643383279502884e+00; }
+template <> inline constexpr float pi() { return 3.141592653589793238462643383279502884e+00F; }
+template <> inline constexpr double pi() { return 3.141592653589793238462643383279502884e+00; }
+template <> inline constexpr long double pi() { return 3.141592653589793238462643383279502884e+00L; }
 
-namespace math {
 
 /// @brief Return the unary negation of the given value.
 /// @note A negative<T>() specialization must be defined for each ValueType T
 /// for which unary negation is not defined.
-template<typename T> inline T negative(const T& val) { return T(-val); }
+template<typename T> inline T negative(const T& val)
+{
+// disable unary minus on unsigned warning
+#if defined(_MSC_VER)
+#pragma warning(push)
+#pragma warning(disable:4146)
+#endif
+    return T(-val);
+#if defined(_MSC_VER)
+#pragma warning(pop)
+#endif
+}
 /// Return the negation of the given boolean.
 template<> inline bool negative(const bool& val) { return !val; }
 /// Return the "negation" of the given string.
@@ -250,7 +280,7 @@ ClampTest01(Type &x)
     return true;
 }
 
-/// @brief Return 0 if @a x < @a 0, 1 if @a x > 1 or else @f$(3-2x)x^2@f$.
+/// @brief Return 0 if @a x < @a 0, 1 if @a x > 1 or else (3 &minus; 2 @a x) @a x&sup2;.
 template<typename Type>
 inline Type
 SmoothUnitStep(Type x)
@@ -258,8 +288,8 @@ SmoothUnitStep(Type x)
     return x > 0 ? x < 1 ? (3-2*x)*x*x : Type(1) : Type(0);
 }
 
-/// @brief Return 0 if @a x < @a min, 1 if @a x > @a max or else @f$(3-2t)t^2@f$,
-/// where @f$t = (x-min)/(max-min)@f$.
+/// @brief Return 0 if @a x < @a min, 1 if @a x > @a max or else (3 &minus; 2 @a t) @a t&sup2;,
+/// where @a t = (@a x &minus; @a min)/(@a max &minus; @a min).
 template<typename Type>
 inline Type
 SmoothUnitStep(Type x, Type min, Type max)
@@ -289,10 +319,10 @@ inline long double Abs(long double x) { return std::fabs(x); }
 inline uint32_t Abs(uint32_t i) { return i; }
 inline uint64_t Abs(uint64_t i) { return i; }
 inline bool Abs(bool b) { return b; }
-// On OSX size_t and uint64_t are different types
-#if defined(__APPLE__) || defined(MACOSX)
-inline size_t Abs(size_t i) { return i; }
-#endif
+// On systems like macOS and FreeBSD, size_t and uint64_t are different types
+template <typename T>
+inline typename std::enable_if<std::is_same<T, size_t>::value, T>::type
+Abs(T i) { return i; }
 //@}
 
 
@@ -337,16 +367,48 @@ template<typename Type>
 inline bool
 isNegative(const Type& x) { return x < zeroVal<Type>(); }
 
-/// Return @c false, since @c bool values are never less than zero.
+// Return false, since bool values are never less than zero.
 template<> inline bool isNegative<bool>(const bool&) { return false; }
 
 
+/// Return @c true if @a x is finite.
+inline bool
+isFinite(const float x) { return std::isfinite(x); }
+
 /// Return @c true if @a x is finite.
 template<typename Type, typename std::enable_if<std::is_arithmetic<Type>::value, int>::type = 0>
 inline bool
-isFinite(const Type& x) { return std::isfinite(x); }
+isFinite(const Type& x) { return std::isfinite(static_cast<double>(x)); }
 
 
+/// Return @c true if @a x is an infinity value (either positive infinity or negative infinity).
+inline bool
+isInfinite(const float x) { return std::isinf(x); }
+
+/// Return @c true if @a x is an infinity value (either positive infinity or negative infinity).
+template<typename Type, typename std::enable_if<std::is_arithmetic<Type>::value, int>::type = 0>
+inline bool
+isInfinite(const Type& x) { return std::isinf(static_cast<double>(x)); }
+
+
+/// Return @c true if @a x is a NaN (Not-A-Number) value.
+inline bool
+isNan(const float x) { return std::isnan(x); }
+
+/// Return @c true if @a x is a NaN (Not-A-Number) value.
+template<typename Type, typename std::enable_if<std::is_arithmetic<Type>::value, int>::type = 0>
+inline bool
+isNan(const Type& x) { return std::isnan(static_cast<double>(x)); }
+
+
+/// Return @c true if @a a is equal to @a b to within the given tolerance.
+template<typename Type>
+inline bool
+isApproxEqual(const Type& a, const Type& b, const Type& tolerance)
+{
+    return !cwiseGreaterThan(Abs(a - b), tolerance);
+}
+
 /// @brief Return @c true if @a a is equal to @a b to within
 /// the default floating-point comparison tolerance.
 template<typename Type>
@@ -354,16 +416,7 @@ inline bool
 isApproxEqual(const Type& a, const Type& b)
 {
     const Type tolerance = Type(zeroVal<Type>() + Tolerance<Type>::value());
-    return !(Abs(a - b) > tolerance);
-}
-
-
-/// Return @c true if @a a is equal to @a b to within the given tolerance.
-template<typename Type>
-inline bool
-isApproxEqual(const Type& a, const Type& b, const Type& tolerance)
-{
-    return !(Abs(a - b) > tolerance);
+    return isApproxEqual(a, b, tolerance);
 }
 
 #define OPENVDB_EXACT_IS_APPROX_EQUAL(T) \
@@ -493,19 +546,19 @@ isUlpsEqual(const float aLeft, const float aRight, const int32_t aUnitsInLastPla
 
 // ==========> Pow <==================
 
-/// Return @f$ x^2 @f$.
+/// Return @a x<sup>2</sup>.
 template<typename Type>
 inline Type Pow2(Type x) { return x*x; }
 
-/// Return @f$ x^3 @f$.
+/// Return @a x<sup>3</sup>.
 template<typename Type>
 inline Type Pow3(Type x) { return x*x*x; }
 
-/// Return @f$ x^4 @f$.
+/// Return @a x<sup>4</sup>.
 template<typename Type>
 inline Type Pow4(Type x) { return Pow2(Pow2(x)); }
 
-/// Return @f$ x^n @f$.
+/// Return @a x<sup>n</sup>.
 template<typename Type>
 Type
 Pow(Type x, int n)
@@ -520,7 +573,7 @@ Pow(Type x, int n)
 }
 
 //@{
-/// Return @f$ b^e @f$.
+/// Return @a b<sup>e</sup>.
 inline float
 Pow(float b, float e)
 {
@@ -655,14 +708,14 @@ Min(const Type& a, const Type& b, const Type& c, const Type& d,
 
 // ============> Exp <==================
 
-/// Return @f$ e^x @f$.
+/// Return @a e<sup>x</sup>.
 template<typename Type>
 inline Type Exp(const Type& x) { return std::exp(x); }
 
 // ============> Sin <==================
 
 //@{
-/// Return @f$ sin(x) @f$.
+/// Return sin @a x.
 inline float Sin(const float& x) { return std::sin(x); }
 
 inline double Sin(const double& x) { return std::sin(x); }
@@ -671,7 +724,7 @@ inline double Sin(const double& x) { return std::sin(x); }
 // ============> Cos <==================
 
 //@{
-/// Return @f$ cos(x) @f$.
+/// Return cos @a x.
 inline float Cos(const float& x) { return std::cos(x); }
 
 inline double Cos(const double& x) { return std::cos(x); }
@@ -819,11 +872,10 @@ template<typename Type>
 inline Type
 Truncate(Type x, unsigned int digits)
 {
-    Type tenth = Pow(10,digits);
+    Type tenth = static_cast<Type>(Pow(size_t(10), digits));
     return RoundDown(x*tenth+0.5)/tenth;
 }
 
-
 ////////////////////////////////////////
 
 
@@ -885,10 +937,7 @@ template<typename Vec3T>
 size_t
 MinIndex(const Vec3T& v)
 {
-#ifndef _MSC_VER // Visual C++ doesn't guarantee thread-safe initialization of local statics
-    static
-#endif
-    const size_t hashTable[8] = { 2, 1, 9, 1, 2, 9, 0, 0 };//9 is a dummy value
+    static const size_t hashTable[8] = { 2, 1, 9, 1, 2, 9, 0, 0 };//9 is a dummy value
     const size_t hashKey =
         ((v[0] < v[1]) << 2) + ((v[0] < v[2]) << 1) + (v[1] < v[2]);// ?*4+?*2+?*1
     return hashTable[hashKey];
@@ -906,10 +955,7 @@ template<typename Vec3T>
 size_t
 MaxIndex(const Vec3T& v)
 {
-#ifndef _MSC_VER // Visual C++ doesn't guarantee thread-safe initialization of local statics
-    static
-#endif
-    const size_t hashTable[8] = { 2, 1, 9, 1, 2, 9, 0, 0 };//9 is a dummy value
+    static const size_t hashTable[8] = { 2, 1, 9, 1, 2, 9, 0, 0 };//9 is a dummy value
     const size_t hashKey =
         ((v[0] > v[1]) << 2) + ((v[0] > v[2]) << 1) + (v[1] > v[2]);// ?*4+?*2+?*1
     return hashTable[hashKey];
@@ -920,7 +966,3 @@ MaxIndex(const Vec3T& v)
 } // namespace openvdb
 
 #endif // OPENVDB_MATH_MATH_HAS_BEEN_INCLUDED
-
-// Copyright (c) 2012-2017 DreamWorks Animation LLC
-// All rights reserved. This software is distributed under the
-// Mozilla Public License 2.0 ( http://www.mozilla.org/MPL/2.0/ )
