@@ -21,7 +21,7 @@
 #include "StreamCompression.h"
 
 #include <tbb/spin_mutex.h>
-#include <tbb/atomic.h>
+#include <atomic>
 
 #include <memory>
 #include <mutex>
@@ -52,7 +52,7 @@ floatingPointToFixedPoint(const FloatT s)
     static_assert(std::is_unsigned<IntegerT>::value, "IntegerT must be unsigned");
     if (FloatT(0.0) > s) return std::numeric_limits<IntegerT>::min();
     else if (FloatT(1.0) <= s) return std::numeric_limits<IntegerT>::max();
-    return IntegerT(std::floor(s * FloatT(std::numeric_limits<IntegerT>::max())));
+    return IntegerT(s * FloatT(std::numeric_limits<IntegerT>::max()));
 }
 
 
@@ -101,7 +101,6 @@ public:
     enum Flag {
         TRANSIENT = 0x1,            /// by default not written to disk
         HIDDEN = 0x2,               /// hidden from UIs or iterators
-        OUTOFCORE = 0x4,            /// data not yet loaded from disk (deprecated flag as of ABI=5)
         CONSTANTSTRIDE = 0x8,       /// stride size does not vary in the array
         STREAMING = 0x10,           /// streaming mode collapses attributes when first accessed
         PARTIALREAD = 0x20          /// data has been partially read (compressed bytes is used)
@@ -126,38 +125,28 @@ public:
     using Ptr           = std::shared_ptr<AttributeArray>;
     using ConstPtr      = std::shared_ptr<const AttributeArray>;
 
-    using FactoryMethod = Ptr (*)(Index, Index, bool);
+    using FactoryMethod = Ptr (*)(Index, Index, bool, const Metadata*);
 
     template <typename ValueType, typename CodecType> friend class AttributeHandle;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     AttributeArray(): mPageHandle() { mOutOfCore = 0; }
-#else
-    AttributeArray(): mPageHandle() {}
-#endif
     virtual ~AttributeArray()
     {
         // if this AttributeArray has been partially read, zero the compressed bytes,
         // so the page handle won't attempt to clean up invalid memory
         if (mFlags & PARTIALREAD)       mCompressedBytes = 0;
     }
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     AttributeArray(const AttributeArray& rhs);
     AttributeArray& operator=(const AttributeArray& rhs);
-#else
-    AttributeArray(const AttributeArray&) = default;
-    AttributeArray& operator=(const AttributeArray&) = default;
-#endif
-    AttributeArray(AttributeArray&&) = default;
-    AttributeArray& operator=(AttributeArray&&) = default;
+    AttributeArray(AttributeArray&&) = delete;
+    AttributeArray& operator=(AttributeArray&&) = delete;
 
     /// Return a copy of this attribute.
     virtual AttributeArray::Ptr copy() const = 0;
 
     /// Return a copy of this attribute.
-    /// @deprecated In-memory compression no longer supported, use AttributeArray::copy() instead.
 #ifndef _MSC_VER
-    OPENVDB_DEPRECATED
+    OPENVDB_DEPRECATED_MESSAGE("In-memory compression no longer supported, use AttributeArray::copy() instead")
 #endif
     virtual AttributeArray::Ptr copyUncompressed() const = 0;
 
@@ -173,7 +162,6 @@ public:
     /// @note This counts each data element in a strided array
     virtual Index dataSize() const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return the name of the value type of a single element in this array (e.g., "float" or "vec3d").
     virtual Name valueType() const = 0;
 
@@ -202,16 +190,26 @@ public:
 
     /// Return @c true if the value type is a matrix
     virtual bool valueTypeIsMatrix() const = 0;
-#endif
 
     /// Return the number of bytes of memory used by this attribute.
     virtual size_t memUsage() const = 0;
 
+#if OPENVDB_ABI_VERSION_NUMBER >= 10
+    /// Return the number of bytes of memory used by this attribute array once it
+    /// has been deserialized (this may be different to memUsage() if delay-loading
+    /// is in use). Note that this method does NOT consider the fact that a
+    /// uniform attribute could be expanded and only deals with delay-loading.
+    virtual size_t memUsageIfLoaded() const = 0;
+#endif
+
     /// Create a new attribute array of the given (registered) type, length and stride.
     /// @details If @a lock is non-null, the AttributeArray registry mutex
     /// has already been locked
     static Ptr create(const NamePair& type, Index length, Index stride = 1,
-        bool constantStride = true, const ScopedRegistryLock* lock = nullptr);
+        bool constantStride = true,
+        const Metadata* metadata = nullptr,
+        const ScopedRegistryLock* lock = nullptr);
+
     /// Return @c true if the given attribute type name is registered.
     static bool isRegistered(const NamePair& type, const ScopedRegistryLock* lock = nullptr);
     /// Clear the attribute type registry.
@@ -228,16 +226,12 @@ public:
     bool hasValueType() const { return this->type().first == typeNameAsString<ValueType>(); }
 
     /// @brief Set value at given index @a n from @a sourceIndex of another @a sourceArray.
-    /// @deprecated From ABI 6 on, use copyValues() with source-target index pairs.
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     // Windows does not allow base classes to be easily deprecated.
 #ifndef _MSC_VER
-    OPENVDB_DEPRECATED
-#endif
+    OPENVDB_DEPRECATED_MESSAGE("Use copyValues() with source-target index pairs")
 #endif
     virtual void set(const Index n, const AttributeArray& sourceArray, const Index sourceIndex) = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// @brief Copy values into this array from a source array to a target array
     /// as referenced by an iterator.
     /// @details Iterators must adhere to the ForwardIterator interface described
@@ -268,7 +262,6 @@ public:
     /// concurrently modified by another thread and that the source array is also not modified.
     template<typename IterT>
     void copyValues(const AttributeArray& sourceArray, const IterT& iter, bool compact = true);
-#endif
 
     /// Return @c true if this array is stored as a single uniform value.
     virtual bool isUniform() const = 0;
@@ -280,19 +273,14 @@ public:
     /// Compact the existing array to become uniform if all values are identical
     virtual bool compact() = 0;
 
-    /// @deprecated Previously this returned @c true if the array was compressed,
-    /// now it always returns @c false.
-    OPENVDB_DEPRECATED bool isCompressed() const { return false; }
-    /// @deprecated Previously this compressed the attribute array, now it does nothing.
     // Windows does not allow base classes to be deprecated
 #ifndef _MSC_VER
-    OPENVDB_DEPRECATED
+    OPENVDB_DEPRECATED_MESSAGE("Previously this compressed the attribute array, now it does nothing")
 #endif
     virtual bool compress() = 0;
-    /// @deprecated Previously this uncompressed the attribute array, now it does nothing.
     // Windows does not allow base classes to be deprecated
 #ifndef _MSC_VER
-    OPENVDB_DEPRECATED
+    OPENVDB_DEPRECATED_MESSAGE("Previously this uncompressed the attribute array, now it does nothing")
 #endif
     virtual bool decompress() = 0;
 
@@ -355,10 +343,8 @@ public:
     /// Ensures all data is in-core
     virtual void loadData() const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return @c true if all data has been loaded
     virtual bool isDataLoaded() const = 0;
-#endif
 
     /// Check the compressed bytes and flags. If they are equal, perform a deeper
     /// comparison check necessary on the inherited types (TypedAttributeArray)
@@ -366,6 +352,11 @@ public:
     bool operator==(const AttributeArray& other) const;
     bool operator!=(const AttributeArray& other) const { return !this->operator==(other); }
 
+#if OPENVDB_ABI_VERSION_NUMBER >= 9
+    /// Indirect virtual function to retrieve the data buffer cast to a char byte array
+    const char* constDataAsByteArray() const { return this->dataAsByteArray(); }
+#endif
+
 private:
     friend class ::TestAttributeArray;
 
@@ -373,7 +364,6 @@ private:
     /// comparisons on inherited types
     virtual bool isEqual(const AttributeArray& other) const = 0;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Virtual function to retrieve the data buffer cast to a char byte array
     virtual char* dataAsByteArray() = 0;
     virtual const char* dataAsByteArray() const = 0;
@@ -382,12 +372,9 @@ private:
     template <typename IterT>
     void doCopyValues(const AttributeArray& sourceArray, const IterT& iter,
         bool rangeChecking = true);
-#endif
 
 protected:
-#if OPENVDB_ABI_VERSION_NUMBER >= 7
     AttributeArray(const AttributeArray& rhs, const tbb::spin_mutex::scoped_lock&);
-#endif
 
     /// @brief Specify whether this attribute has a constant stride or not.
     void setConstantStride(bool state);
@@ -402,30 +389,16 @@ protected:
     static void unregisterType(const NamePair& type,
         const ScopedRegistryLock* lock = nullptr);
 
-#if OPENVDB_ABI_VERSION_NUMBER < 6
-
-    size_t mCompressedBytes = 0;
-    uint8_t mFlags = 0;
-    uint8_t mUsePagedRead = 0;
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
-    tbb::atomic<Index32> mOutOfCore; // interpreted as bool
-#endif
-    compression::PageHandle::Ptr mPageHandle;
-
-#else // #if OPENVDB_ABI_VERSION_NUMBER < 6
-
     bool mIsUniform = true;
     mutable tbb::spin_mutex mMutex;
     uint8_t mFlags = 0;
     uint8_t mUsePagedRead = 0;
-    tbb::atomic<Index32> mOutOfCore; // interpreted as bool
+    std::atomic<Index32> mOutOfCore; // interpreted as bool
     /// used for out-of-core, paged reading
     union {
         compression::PageHandle::Ptr mPageHandle;
-        size_t mCompressedBytes; // as of ABI=6, this data is packed together to save memory
+        size_t mCompressedBytes;
     };
-
-#endif
 }; // class AttributeArray
 
 
@@ -460,7 +433,7 @@ struct AttributeArray::Accessor : public AttributeArray::AccessorBase
 namespace attribute_traits
 {
     template <typename T> struct TruncateTrait { };
-    template <> struct TruncateTrait<float> { using Type = half; };
+    template <> struct TruncateTrait<float> { using Type = math::half; };
     template <> struct TruncateTrait<int> { using Type = short; };
 
     template <typename T> struct TruncateTrait<math::Vec3<T>> {
@@ -562,11 +535,7 @@ struct UnitVecCodec
 /// Typed class for storing attribute data
 
 template<typename ValueType_, typename Codec_ = NullCodec>
-#if OPENVDB_ABI_VERSION_NUMBER >= 6 // for ABI=6, class is final to allow for de-virtualization
 class TypedAttributeArray final: public AttributeArray
-#else
-class TypedAttributeArray: public AttributeArray
-#endif
 {
 public:
     using Ptr           = std::shared_ptr<TypedAttributeArray>;
@@ -581,7 +550,7 @@ public:
     /// Default constructor, always constructs a uniform attribute.
     explicit TypedAttributeArray(Index n = 1, Index strideOrTotalSize = 1, bool constantStride = true,
         const ValueType& uniformValue = zeroVal<ValueType>());
-#if OPENVDB_ABI_VERSION_NUMBER >= 7
+
     /// Deep copy constructor.
     /// @note This method is thread-safe (as of ABI=7) for concurrently reading from the
     /// source attribute array while being deep-copied. Specifically, this means that the
@@ -590,14 +559,9 @@ public:
     /// It is not thread-safe for write.
     TypedAttributeArray(const TypedAttributeArray&);
     /// Deep copy constructor.
-    /// @deprecated Use copy-constructor without unused bool parameter
-    OPENVDB_DEPRECATED TypedAttributeArray(const TypedAttributeArray&, bool /*unused*/);
-#else
-    /// Deep copy constructor.
-    /// @note This method is not thread-safe for reading or writing, use
-    /// TypedAttributeArray::copy() to ensure thread-safety when reading concurrently.
-    TypedAttributeArray(const TypedAttributeArray&, bool uncompress = false);
-#endif
+    OPENVDB_DEPRECATED_MESSAGE("Use copy-constructor without unused bool parameter")
+    TypedAttributeArray(const TypedAttributeArray&, bool /*unused*/);
+
     /// Deep copy assignment operator.
     /// @note this operator is thread-safe.
     TypedAttributeArray& operator=(const TypedAttributeArray&);
@@ -612,12 +576,14 @@ public:
     /// @note This method is thread-safe.
     AttributeArray::Ptr copy() const override;
 
-    /// Return an uncompressed copy of this attribute (will just return a copy if not compressed).
+    /// Return a copy of this attribute.
     /// @note This method is thread-safe.
-    OPENVDB_DEPRECATED AttributeArray::Ptr copyUncompressed() const override;
+    OPENVDB_DEPRECATED_MESSAGE("In-memory compression no longer supported, use AttributeArray::copy() instead")
+    AttributeArray::Ptr copyUncompressed() const override;
 
     /// Return a new attribute array of the given length @a n and @a stride with uniform value zero.
-    static Ptr create(Index n, Index strideOrTotalSize = 1, bool constantStride = true);
+    static Ptr create(Index n, Index strideOrTotalSize = 1, bool constantStride = true,
+        const Metadata* metadata = nullptr);
 
     /// Cast an AttributeArray to TypedAttributeArray<T>
     static TypedAttributeArray& cast(AttributeArray& attributeArray);
@@ -649,7 +615,6 @@ public:
         return hasConstantStride() ? mSize * mStrideOrTotalSize : mStrideOrTotalSize;
     }
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return the name of the value type of a single element in this array (e.g., "float" or "vec3d").
     Name valueType() const override { return typeNameAsString<ValueType>(); }
 
@@ -677,41 +642,46 @@ public:
 
     /// Return @c true if the value type is a matrix
     bool valueTypeIsMatrix() const override;
-#endif
 
     /// Return the number of bytes of memory used by this attribute.
     size_t memUsage() const override;
 
-    /// Return the value at index @a n (assumes uncompressed and in-core)
+#if OPENVDB_ABI_VERSION_NUMBER >= 10
+    /// Return the number of bytes of memory used by this attribute array once it
+    /// has been deserialized (this may be different to memUsage() if delay-loading
+    /// is in use). Note that this method does NOT consider the fact that a
+    /// uniform attribute could be expanded and only deals with delay-loading.
+    size_t memUsageIfLoaded() const override;
+#endif
+
+    /// Return the value at index @a n (assumes in-core)
     ValueType getUnsafe(Index n) const;
     /// Return the value at index @a n
     ValueType get(Index n) const;
-    /// Return the @a value at index @a n (assumes uncompressed and in-core)
+    /// Return the @a value at index @a n (assumes in-core)
     template<typename T> void getUnsafe(Index n, T& value) const;
     /// Return the @a value at index @a n
     template<typename T> void get(Index n, T& value) const;
 
     /// Non-member equivalent to getUnsafe() that static_casts array to this TypedAttributeArray
-    /// (assumes uncompressed and in-core)
+    /// (assumes in-core)
     static ValueType getUnsafe(const AttributeArray* array, const Index n);
 
-    /// Set @a value at the given index @a n (assumes uncompressed and in-core)
+    /// Set @a value at the given index @a n (assumes in-core)
     void setUnsafe(Index n, const ValueType& value);
     /// Set @a value at the given index @a n
     void set(Index n, const ValueType& value);
-    /// Set @a value at the given index @a n (assumes uncompressed and in-core)
+    /// Set @a value at the given index @a n (assumes in-core)
     template<typename T> void setUnsafe(Index n, const T& value);
     /// Set @a value at the given index @a n
     template<typename T> void set(Index n, const T& value);
 
     /// Non-member equivalent to setUnsafe() that static_casts array to this TypedAttributeArray
-    /// (assumes uncompressed and in-core)
+    /// (assumes in-core)
     static void setUnsafe(AttributeArray* array, const Index n, const ValueType& value);
 
     /// Set value at given index @a n from @a sourceIndex of another @a sourceArray
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
-    OPENVDB_DEPRECATED
-#endif
+    OPENVDB_DEPRECATED_MESSAGE("Use copyValues() with source-target index pairs")
     void set(const Index n, const AttributeArray& sourceArray, const Index sourceIndex) override;
 
     /// Return @c true if this array is stored as a single uniform value.
@@ -737,9 +707,11 @@ public:
     static void fill(AttributeArray* array, const ValueType& value);
 
     /// Compress the attribute array.
-    OPENVDB_DEPRECATED bool compress() override;
+    OPENVDB_DEPRECATED_MESSAGE("Previously this compressed the attribute array, now it does nothing")
+    bool compress() override;
     /// Uncompress the attribute array.
-    OPENVDB_DEPRECATED bool decompress() override;
+    OPENVDB_DEPRECATED_MESSAGE("Previously this uncompressed the attribute array, now it does nothing")
+    bool decompress() override;
 
     /// Read attribute data from a stream.
     void read(std::istream&) override;
@@ -778,9 +750,12 @@ public:
     /// Ensures all data is in-core
     void loadData() const override;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Return @c true if all data has been loaded
     bool isDataLoaded() const override;
+
+#if OPENVDB_ABI_VERSION_NUMBER >= 9
+    /// Return the raw data buffer
+    inline const StorageType* constData() const { return this->data(); }
 #endif
 
 protected:
@@ -796,14 +771,12 @@ protected:
 private:
     friend class ::TestAttributeArray;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 7
     TypedAttributeArray(const TypedAttributeArray&, const tbb::spin_mutex::scoped_lock&);
-#endif
 
     /// Load data from memory-mapped file.
     inline void doLoad() const;
     /// Load data from memory-mapped file (unsafe as this function is not protected by a mutex).
-    /// @param compression if true, loading previously compressed data will re-compressed it
+    /// @param compression parameter no longer used
     inline void doLoadUnsafe(const bool compression = true) const;
     /// Compress in-core data assuming mutex is locked
     inline bool compressUnsafe();
@@ -814,29 +787,24 @@ private:
     /// Compare the this data to another attribute array. Used by the base class comparison operator
     bool isEqual(const AttributeArray& other) const override;
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
     /// Virtual function to retrieve the data buffer from the derived class cast to a char byte array
     char* dataAsByteArray() override;
     const char* dataAsByteArray() const override;
-#endif
 
     size_t arrayMemUsage() const;
     void allocate();
     void deallocate();
 
     /// Helper function for use with registerType()
-    static AttributeArray::Ptr factory(Index n, Index strideOrTotalSize, bool constantStride) {
-        return TypedAttributeArray::create(n, strideOrTotalSize, constantStride);
+    static AttributeArray::Ptr factory(Index n, Index strideOrTotalSize, bool constantStride,
+        const Metadata* metadata) {
+        return TypedAttributeArray::create(n, strideOrTotalSize, constantStride, metadata);
     }
 
     static std::unique_ptr<const NamePair> sTypeName;
     std::unique_ptr<StorageType[]>      mData;
     Index                               mSize;
     Index                               mStrideOrTotalSize;
-#if OPENVDB_ABI_VERSION_NUMBER < 6 // as of ABI=6, this data lives in the base class to reduce memory
-    bool                                mIsUniform = true;
-    mutable tbb::spin_mutex             mMutex;
-#endif
 }; // class TypedAttributeArray
 
 
@@ -1045,8 +1013,6 @@ UnitVecCodec::encode(const math::Vec3<T>& val, StorageType& data)
 
 // AttributeArray implementation
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
-
 template <typename IterT>
 void AttributeArray::doCopyValues(const AttributeArray& sourceArray, const IterT& iter,
     bool rangeChecking/*=true*/)
@@ -1135,7 +1101,6 @@ void AttributeArray::copyValues(const AttributeArray& sourceArray, const IterT&
         this->compact();
     }
 }
-#endif
 
 
 ////////////////////////////////////////
@@ -1174,7 +1139,6 @@ TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 7
 template<typename ValueType_, typename Codec_>
 TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs)
     : TypedAttributeArray(rhs, tbb::spin_mutex::scoped_lock(rhs.mMutex))
@@ -1186,20 +1150,12 @@ template<typename ValueType_, typename Codec_>
 TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs,
     const tbb::spin_mutex::scoped_lock& lock)
     : AttributeArray(rhs, lock)
-#else
-template<typename ValueType_, typename Codec_>
-TypedAttributeArray<ValueType_, Codec_>::TypedAttributeArray(const TypedAttributeArray& rhs, bool)
-    : AttributeArray(rhs)
-#endif
     , mSize(rhs.mSize)
     , mStrideOrTotalSize(rhs.mStrideOrTotalSize)
-#if OPENVDB_ABI_VERSION_NUMBER < 6
-    , mIsUniform(rhs.mIsUniform)
-#endif
 {
     if (this->validData()) {
         this->allocate();
-        std::memcpy(this->data(), rhs.data(), this->arrayMemUsage());
+        std::memcpy(static_cast<void*>(this->data()), rhs.data(), this->arrayMemUsage());
     }
 }
 
@@ -1223,9 +1179,11 @@ TypedAttributeArray<ValueType_, Codec_>::operator=(const TypedAttributeArray& rh
 
         if (this->validData()) {
             this->allocate();
-            std::memcpy(this->newDataAsByteArray(), rhs.newDataAsByteArray(), this->arrayMemUsage());
+            std::memcpy(static_cast<void*>(this->data()), rhs.data(), this->arrayMemUsage());
         }
     }
+
+    return *this;
 }
 
 
@@ -1268,9 +1226,14 @@ TypedAttributeArray<ValueType_, Codec_>::unregisterType()
 
 template<typename ValueType_, typename Codec_>
 inline typename TypedAttributeArray<ValueType_, Codec_>::Ptr
-TypedAttributeArray<ValueType_, Codec_>::create(Index n, Index stride, bool constantStride)
+TypedAttributeArray<ValueType_, Codec_>::create(Index n, Index stride, bool constantStride,
+    const Metadata* metadata)
 {
-    return Ptr(new TypedAttributeArray(n, stride, constantStride));
+    const TypedMetadata<ValueType>* typedMetadata = metadata ?
+        dynamic_cast<const TypedMetadata<ValueType>*>(metadata) : nullptr;
+
+    return Ptr(new TypedAttributeArray(n, stride, constantStride,
+        typedMetadata ? typedMetadata->value() : zeroVal<ValueType>()));
 }
 
 template<typename ValueType_, typename Codec_>
@@ -1297,9 +1260,6 @@ template<typename ValueType_, typename Codec_>
 AttributeArray::Ptr
 TypedAttributeArray<ValueType_, Codec_>::copy() const
 {
-#if OPENVDB_ABI_VERSION_NUMBER < 7
-    tbb::spin_mutex::scoped_lock lock(mMutex);
-#endif
     return AttributeArray::Ptr(new TypedAttributeArray<ValueType, Codec>(*this));
 }
 
@@ -1351,7 +1311,6 @@ TypedAttributeArray<ValueType_, Codec_>::deallocate()
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::valueTypeIsFloatingPoint() const
@@ -1368,7 +1327,7 @@ TypedAttributeArray<ValueType_, Codec_>::valueTypeIsFloatingPoint() const
     using ElementT = typename VecTraits<ValueType>::ElementType;
 
     // half is not defined as float point as expected, so explicitly handle it
-    return std::is_floating_point<ElementT>::value || std::is_same<half, ElementT>::value;
+    return std::is_floating_point<ElementT>::value || std::is_same<math::half, ElementT>::value;
 }
 
 
@@ -1377,7 +1336,7 @@ bool
 TypedAttributeArray<ValueType_, Codec_>::valueTypeIsClass() const
 {
     // half is not defined as a non-class type as expected, so explicitly exclude it
-    return std::is_class<ValueType>::value && !std::is_same<half, ValueType>::value;
+    return std::is_class<ValueType>::value && !std::is_same<math::half, ValueType>::value;
 }
 
 
@@ -1405,7 +1364,6 @@ TypedAttributeArray<ValueType_, Codec_>::valueTypeIsMatrix() const
     // TODO: improve performance by making this a compile-time check using type traits
     return !this->valueType().compare(0, 3, "mat");
 }
-#endif
 
 
 template<typename ValueType_, typename Codec_>
@@ -1415,6 +1373,15 @@ TypedAttributeArray<ValueType_, Codec_>::memUsage() const
     return sizeof(*this) + (bool(mData) ? this->arrayMemUsage() : 0);
 }
 
+#if OPENVDB_ABI_VERSION_NUMBER >= 10
+template<typename ValueType_, typename Codec_>
+size_t
+TypedAttributeArray<ValueType_, Codec_>::memUsageIfLoaded() const
+{
+    return sizeof(*this) + (mIsUniform ? 1 : this->dataSize()) * sizeof(StorageType);
+}
+#endif
+
 
 template<typename ValueType_, typename Codec_>
 typename TypedAttributeArray<ValueType_, Codec_>::ValueType
@@ -1652,11 +1619,7 @@ template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::isOutOfCore() const
 {
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     return mOutOfCore;
-#else
-    return (mFlags & OUTOFCORE);
-#endif
 }
 
 
@@ -1664,12 +1627,7 @@ template<typename ValueType_, typename Codec_>
 void
 TypedAttributeArray<ValueType_, Codec_>::setOutOfCore(const bool b)
 {
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     mOutOfCore = b;
-#else
-    if (b) mFlags = static_cast<uint8_t>(mFlags | OUTOFCORE);
-    else   mFlags = static_cast<uint8_t>(mFlags & ~OUTOFCORE);
-#endif
 }
 
 
@@ -1697,14 +1655,12 @@ TypedAttributeArray<ValueType_, Codec_>::loadData() const
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 bool
 TypedAttributeArray<ValueType_, Codec_>::isDataLoaded() const
 {
     return !this->isOutOfCore();
 }
-#endif
 
 
 template<typename ValueType_, typename Codec_>
@@ -1842,6 +1798,7 @@ TypedAttributeArray<ValueType_, Codec_>::readPagedBuffers(compression::PagedInpu
     if (!delayLoad) {
         std::unique_ptr<char[]> buffer = mPageHandle->read();
         mData.reset(reinterpret_cast<StorageType*>(buffer.release()));
+        mPageHandle.reset();
     }
 
     // clear page state
@@ -1877,11 +1834,7 @@ TypedAttributeArray<ValueType_, Codec_>::writeMetadata(std::ostream& os, bool ou
         OPENVDB_THROW(IoError, "Cannot write out a partially-read AttributeArray.");
     }
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     uint8_t flags(mFlags);
-#else
-    uint8_t flags(mFlags & uint8_t(~OUTOFCORE));
-#endif
     uint8_t serializationFlags(0);
     Index size(mSize);
     Index stride(mStrideOrTotalSize);
@@ -2016,11 +1969,7 @@ TypedAttributeArray<ValueType_, Codec_>::doLoadUnsafe(const bool /*compression*/
 
     // clear all write and out-of-core flags
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 5
     self->mOutOfCore = false;
-#else
-    self->mFlags &= uint8_t(~OUTOFCORE);
-#endif
 }
 
 
@@ -2029,7 +1978,7 @@ AttributeArray::AccessorBasePtr
 TypedAttributeArray<ValueType_, Codec_>::getAccessor() const
 {
     // use the faster 'unsafe' get and set methods as attribute handles
-    // ensure data is uncompressed and in-core when constructed
+    // ensure data is in-core when constructed
 
     return AccessorBasePtr(new AttributeArray::Accessor<ValueType_>(
         &TypedAttributeArray<ValueType_, Codec_>::getUnsafe,
@@ -2062,7 +2011,6 @@ TypedAttributeArray<ValueType_, Codec_>::isEqual(const AttributeArray& other) co
 }
 
 
-#if OPENVDB_ABI_VERSION_NUMBER >= 6
 template<typename ValueType_, typename Codec_>
 char*
 TypedAttributeArray<ValueType_, Codec_>::dataAsByteArray()
@@ -2077,7 +2025,6 @@ TypedAttributeArray<ValueType_, Codec_>::dataAsByteArray() const
 {
     return reinterpret_cast<const char*>(this->data());
 }
-#endif
 
 
 ////////////////////////////////////////
